
from tqdm import tqdm
import os
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


def search_through_hca_metadata_for_value(tree, current_key=None, key=None, value=None, project_key=None,
                                          search_type="full"):
    if search_type not in ["full", "partial"]:
        raise ValueError("The argument search_type must be either of partial or full.")
    if search_type == "full":
        if isinstance(tree, list):
            search_condition_list = value in filter(None, tree)
        elif not isinstance(tree, list) and not isinstance(tree, dict):
            search_condition_top = tree == value
    elif search_type == "partial":
        if isinstance(tree, list):
            search_condition_list = any(value.lower() in s for s in filter(None, tree) if isinstance(s, str)) or \
                                    any(value.capitalize() in s for s in filter(None, tree) if isinstance(s, str)) or \
                                    any(value.upper() in s for s in filter(None, tree) if isinstance(s, str))
        elif not isinstance(tree, list) and not isinstance(tree, dict):
            search_condition_top = tree is not None and tree is not False and isinstance(tree, str) and \
                                   (value.lower() in tree or value.capitalize() in tree or value.upper() in tree)

    if not isinstance(tree, list) and not isinstance(tree, dict) and key == current_key and search_condition_top:
        yield project_key
        pass
    elif isinstance(tree, list) and key == current_key and search_condition_list:
        yield project_key
        pass
    # case 1: if it is still a dictionary, recursively go through all entries
    elif isinstance(tree, dict):
        for sub_key, sub_value in tree.items():
            yield from search_through_hca_metadata_for_value(sub_value, sub_key, key, value, project_key, search_type)
    # case 2: if it is a list, evaluate the elem
    elif isinstance(tree, list):
        # case 2.1: if any of the elems are a dictionary, recursively go through them
        for element in tree:
            yield from search_through_hca_metadata_for_value(element, current_key, key, value, project_key, search_type)


def iterate_matrices_tree(tree, keys=()):
    if isinstance(tree, dict):
        for k, v in tree.items():
            yield from iterate_matrices_tree(v, keys=(*keys, k))
    elif isinstance(tree, list):
        for file in tree:
            yield keys, file
    else:
        assert False


def bulk_download_files(endpoint_url, save_location, catalog='dcp22'):
    if not os.path.exists(save_location):
        os.mkdir(save_location)
    response = requests.get(endpoint_url, params={'catalog': catalog})
    response.raise_for_status()

    response_json = response.json()
    project = response_json['projects'][0]

    file_urls = set()
    for key in ('matrices', 'contributedAnalyses'):
        tree = project[key]
        for path, file_info in iterate_matrices_tree(tree):
            url = file_info['url']
            if url not in file_urls:
                dest_path = os.path.join(save_location, file_info['name'])
                print(file_info['name'])
                if ".h5ad" in file_info['name'] and not os.path.isfile(dest_path):
                    download_file(url, dest_path)
                    file_urls.add(url)
    print('Downloads Complete.')


def collect_unique_hca_metadata_fields(tree, tree_key=None):

    # case 1: if it is still a dictionary, recursively go through all entries
    if isinstance(tree, dict):
        for key, value in tree.items():
            yield from collect_unique_hca_metadata_fields(value, key)
    # case 2: if it is a list, evaluate the elem
    elif isinstance(tree, list):
        # case 2.1: if any of the elems are a dictionary, recursively go through them
        if any(isinstance(element, dict) for element in tree):
            for element in tree:
                if isinstance(element, dict):
                    for key, value in element.items():
                        yield from collect_unique_hca_metadata_fields(value, key)
                elif isinstance(element, list):
                    for sub_elem in element:
                        yield from collect_unique_hca_metadata_fields(sub_elem, element)
                else:
                    yield from collect_unique_hca_metadata_fields(element, tree_key)
        else:
            for element in tree:
                yield {tree_key: element}
    else:
        yield {tree_key: tree}


def download_file(url, output_path):
    url = url.replace('/fetch', '')  # Work around https://github.com/DataBiosphere/azul/issues/2908

    session = requests.Session()
    retry = Retry(connect=3, backoff_factor=0.5)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    response = session.get(url, stream=True)
    response.raise_for_status()

    total = int(response.headers.get('content-length', 0))
    print(f'Downloading to: {output_path}', flush=True)

    with open(output_path, 'wb') as f:
        with tqdm(total=total, unit='B', unit_scale=True, unit_divisor=1024) as bar:
            for chunk in response.iter_content(chunk_size=1024):
                size = f.write(chunk)
                bar.update(size)



import pytest
from pascrd.api.tabula_sapiens import TabulaSapiensParser, download_tabula_sapiens_dataset
from pascrd.api.human_cell_atlas import HCAParser
from pascrd.utils import collect_unique_hca_metadata_fields
import requests_mock
import os
import mock
import zipfile
from zipfile import BadZipfile
import responses
from unittest import mock
import shutil
import collections


def test_basic_parser_chrome():
    try:
        parser = TabulaSapiensParser(time_delay=0.1)
        parser.collect_datasets()
        assert parser.file_number == len(parser.download_links) == 30
    except ValueError:
        pytest.fail("The parser is not valid.")


def test_basic_parser_firefox():
    try:
        parser = TabulaSapiensParser(browser="firefox", time_delay=0.5)
        parser.collect_datasets()
        assert parser.file_number == len(parser.download_links) == 30
    except ValueError:
        pytest.fail("The parser is not valid.")


def test_invalid_browser():
    with pytest.raises(ValueError):
        TabulaSapiensParser(browser="fake")


def test_basic_hca_parser():
    parser = HCAParser()
    assert len(parser.project_metadata) == 313


def test_basic_hca_parser_reload():
    parser = HCAParser()
    first_metadata = parser.project_metadata
    parser.collect_project_identifiers()
    parser.collect_project_metadata()
    assert len(parser.project_identifiers) == len(parser.project_metadata) == 313
    assert collections.Counter(first_metadata.keys()) == collections.Counter(parser.project_metadata.keys())


@pytest.fixture(scope="function")
def get_tmp_ts_file(tmp_path):
    return str(os.path.join(tmp_path))


@requests_mock.Mocker(kw="mock")
def test_mock_download_tabula_sapiens(get_tmp_ts_file, **kwargs):
    expected_headers = {'Content-Type': 'text/html', 'Content-Length': '1'}
    kwargs["mock"].get('http://test.com', headers=expected_headers)
    download_tabula_sapiens_dataset("fake_key", "http://test.com", get_tmp_ts_file, chunk_size=1, use_unzip=False)

    assert os.path.isfile(os.path.join(get_tmp_ts_file, "fake_key.h5ad.zip"))


@requests_mock.Mocker(kw="mock")
def test_mock_download_tabula_sapiens_zipped(get_tmp_ts_file, **kwargs):
    with pytest.raises(BadZipfile):
        expected_headers = {'Content-Type': 'application/zip', 'Content-Length': '1',
                        'one': 'one', 'two': 'two', 'three': 'three'}
        kwargs["mock"].get('http://test.com', headers=expected_headers)
        download_tabula_sapiens_dataset("fake_key", "http://test.com", get_tmp_ts_file, chunk_size=1, use_unzip=True)
        assert os.path.isfile(os.path.join(get_tmp_ts_file, "fake_key.h5ad.zip"))

    shutil.copy(os.path.join(os.path.dirname(__file__), 'data', 'test.zip'), os.path.join(get_tmp_ts_file,
                                                                                          "fake_key.h5ad.zip"))
    download_tabula_sapiens_dataset("fake_key", "http://test.com", get_tmp_ts_file, chunk_size=1, use_unzip=True)
    assert os.path.isfile(os.path.join(get_tmp_ts_file, "fake_key.h5ad.zip"))


def test_bad_search():
    parser = HCAParser()
    with pytest.raises(ValueError):
        parser.search(search_type="fake")


def test_basic_search():
    parser = HCAParser()
    one_search_union = parser.search({"effectiveOrgan": "blood"})
    two_search_union = parser.search({"effectiveOrgan": "blood", 'institution': 'Broad Institute'})
    assert one_search_union != two_search_union
    assert len(two_search_union) > len(one_search_union)

    breast_parser = parser.search({"effectiveOrgan": "breast", "projectId": "a004b150-1c36-4af6-9bbd-070c06dbc17d"},
                                  search_type="union")
    assert len(breast_parser) == 11

    breast_parser_2 = parser.search({"effectiveOrgan": "breast", "projectId": "a004b150-1c36-4af6-9bbd-070c06dbc17d"},
                                    search_type="intersection")
    assert len(breast_parser_2) == 1
    assert breast_parser != breast_parser_2
    assert len(breast_parser) > len(breast_parser_2)


def test_basic_search_2():
    parser = HCAParser()
    multi_search_union = parser.search({"genusSpecies": "Mus musculus", 'organ': 'brain'}, search_type="intersection")
    assert len(multi_search_union) == 9
    same_search_partial = parser.search({"genusSpecies": "mus", 'organ': 'BRAI'}, search_type="intersection",
                                        match_type="partial")
    assert len(same_search_partial) == len(multi_search_union) == 9
    multi_search_union = parser.search({"genusSpecies": "Mus musculus", 'organ': 'brain',
                                        'organPart': 'skin of body'}, search_type="intersection")
    assert len(multi_search_union) == 1


def test_full_vs_partial_search():
    parser = HCAParser()
    empty_exact = parser.search({"effectiveOrgan": "esophagu"})
    proper_exact = parser.search({"effectiveOrgan": "esophagus"})
    assert len(empty_exact) == 0
    assert len(proper_exact) > 0
    partial_close = parser.search({"organ": "esophagu"}, match_type="partial")
    assert len(partial_close) == len(proper_exact)


def test_multi_search():
    parser = HCAParser()
    one_organ = parser.search({"organ": "esophagus"})
    one_organ_2 = parser.search({"organ": "nose"})
    two_organ = parser.search({"organ": ["esophagus", "nose"]})
    two_organ_off = parser.search({"organ": ["esophagu", "nose", "bloo"]})
    two_organ_partial = parser.search({"organ": ["esophagu", "nose"]}, match_type="partial")
    assert len(two_organ_partial) == len(two_organ)
    assert len(two_organ_off) == len(one_organ_2)
    assert len(two_organ) == len(one_organ) + len(one_organ_2)
    no_overlap = parser.search({"organ": ["esophagus", "nose"]}, search_type="intersection")
    assert len(no_overlap) == 0


def test_basic_collect_search_options():
    parser = HCAParser()
    assert {"organ", "laboratory", "institution", "email"}.issubset(parser.search_options.keys())
    assert "blood" in parser.search_options["organ"]



import urllib.request
import os
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import requests
import logging
import json
from pascrd.utils import search_through_hca_metadata_for_value, collect_unique_hca_metadata_fields
import asyncio
import aiohttp


class HCAParser:
    def __init__(self, repo_directory="https://service.azul.data.humancellatlas.org/index/projects/",
                 session_retries=3, session_backoff=0.5):
        self.process_count = None
        self.directory = repo_directory
        self.session = requests.Session()
        session_retry = Retry(connect=session_retries, backoff_factor=session_backoff)
        session_adapter = HTTPAdapter(max_retries=session_retry)
        self.session.mount('http://', session_adapter)
        self.session.mount('https://', session_adapter)
        self.project_identifiers = {}
        logging.basicConfig(level=logging.INFO)
        self.logger = logging.getLogger()
        self.catalog = None
        if os.path.isfile(str(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                           'data', 'hca.json')))):
            with open(str(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                       'data', 'hca.json')))) as metadata_json:
                self.project_metadata = json.load(metadata_json)
        else:
            self.project_metadata = None

        self.search_results = None
        self.search_options = None
        if self.project_metadata is not None:
            self._collect_search_options()

    def collect_project_identifiers(self):
        with urllib.request.urlopen(self.directory) as project_url:
            data = json.load(project_url)
            json_dict = data['termFacets']['project']
            for elem in json_dict['terms']:
                self.project_identifiers[elem['term']] = elem['projectId'][0] if \
                    isinstance(elem['projectId'], list) else elem['projectId']

    def get_project_json_url(self, project):
        return f'{self.directory}/{project}' if not self.directory.endswith('/') else f'{self.directory}{project}'

    async def get_hca_url(self, identifier, session, verbose=True):
        try:
            url = self.get_project_json_url(identifier)
            async with session.get(url=url, params={'catalog': self.catalog}) as response:
                finding = await response.json()
                self.project_metadata[identifier] = finding
                self.process_count += 1
                if verbose and self.process_count % 10 == 0:
                    self.logger.info(f"Processing dataset {self.process_count} of {len(self.project_identifiers)}")
        except Exception as e:
            self.logger.info("Unable to get url {} due to {}.".format(url, e.__class__))

    async def main(self, query_dict, verbose=True):
        async with aiohttp.ClientSession() as session:
            await asyncio.gather(*[self.get_hca_url(query_value, session, verbose) for query_key, query_value in
                                   query_dict.items()])

    def collect_project_metadata(self, verbose=True, catalog="dcp23", write_local=True):
        self.project_metadata = {}
        self.catalog = catalog
        self.process_count = 0

        asyncio.run(self.main(self.project_identifiers, verbose))

        if write_local:
            with open(str(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                       'data', 'hca.json'))), 'w') as metadata_json:
                json.dump(self.project_metadata, metadata_json)

    def search(self, search_dict=None, search_type="union", match_type="full"):
        if search_type not in ["intersection", "union"]:
            raise ValueError("The argument search_type must be either of intersection or union.")
        if match_type not in ["full", "partial"]:
            raise ValueError("The argument match_type must be either of partial or full.")

        search_results = []
        for key, value in search_dict.items():
            originally_list = isinstance(value, list)
            value = [value] if not originally_list else value
            if not originally_list:
                one_search_results = []
            for sub_search in value:
                if originally_list:
                    one_search_results = []
                for project_key, project_values in self.project_metadata.items():
                    for searched_elem in search_through_hca_metadata_for_value(project_values, key=key,
                                                                               value=sub_search,
                                                                           project_key=project_key,
                                                                           search_type=match_type):
                        if searched_elem not in one_search_results:
                            one_search_results.append(searched_elem)
                    if originally_list:
                        search_results.append(one_search_results)
            if not originally_list:
                search_results.append(one_search_results)
        if search_type == "union":
            found = []
            for sub_list in search_results:
                for elem in sub_list:
                    if elem not in found:
                        found.append(elem)
            return found
        elif search_type == "intersection":
            counts = {}
            for sub_list in search_results:
                for elem in sub_list:
                    if elem not in counts.keys():
                        counts[elem] = 1
                    else:
                        counts[elem] += 1
            return list({k: v for (k, v) in counts.items() if v == len(search_dict)}.keys())

    def _collect_search_options(self):
        self.search_options = {}
        for key, value in self.project_metadata.items():
            for project_elem in collect_unique_hca_metadata_fields(value):
                for sub_key, sub_value in project_elem.items():
                    if sub_key not in self.search_options.keys():
                        self.search_options[sub_key] = [sub_value]
                    else:
                        if sub_value not in self.search_options[sub_key]:
                            self.search_options[sub_key].append(sub_value)
        return self.search_options



import logging
import time
import re
from selenium.webdriver.common.by import By
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.common.exceptions import StaleElementReferenceException, NoSuchElementException
import requests
import os
from tqdm import *
import zipfile


class TabulaSapiensParser:
    def __init__(self, url='https://figshare.com/articles/dataset/Tabula_Sapiens_release_1_0/14267219',
                 time_delay=0.1, browser="chrome"):

        if browser not in ["chrome", "firefox"]:
            raise ValueError("the browser value must be one of: chrome, firefox")

        self.url = url
        self.time_delay = time_delay
        self.browser = browser
        self.options = ChromeOptions() if self.browser == "chrome" else FirefoxOptions()
        self.options.add_argument('--headless')
        self.options.add_argument('--disable-gpu')
        self.driver = webdriver.Chrome(options=self.options) if self.browser == "chrome" else \
            webdriver.Firefox(options=self.options)

        self.driver.get(self.url)
        self.download_links = {}

        try:
            file_counter = self.driver.find_element(By.CLASS_NAME, "_1Xdzb").text
            assert "files" in str(file_counter)
            self.file_number = int(file_counter.split(" files")[0])
        except NoSuchElementException:
            logging.info(f"Verify that the url {url} is a valid link for Tabula Sapiens download links.")

    def _get_links(self):
        time.sleep(self.time_delay)
        try:
            links = self.driver.find_elements(by=By.TAG_NAME, value='a')
            for link in links:
                if "ndownloader" in link.get_attribute('href') and link.get_attribute('href') not in \
                        self.download_links.values():
                    parent_of_parent = link.find_element(By.XPATH, '../..')
                    for elem in parent_of_parent.find_elements(by=By.TAG_NAME, value='span'):
                        if "zip" in elem.get_attribute('title'):
                            self.download_links[re.sub(r'^.*?TS_', '',
                                                       elem.get_attribute('title')).split('.h5ad')[0].title()] = \
                                link.get_attribute('href')
        except StaleElementReferenceException:
            self.time_delay = 1.1 * self.time_delay
            logging.info(f"increasing the rendering delay: {self.time_delay}")
            self._get_links()

    def collect_datasets(self):
        while len(self.download_links) != self.file_number:
            time.sleep(self.time_delay)
            self.driver.find_element(By.XPATH, "//button[.='Next page']").click()
            self._get_links()
        assert len(self.download_links) == self.file_number


def download_tabula_sapiens_dataset(dataset_key: str, dataset_url: str, destination_path: str, chunk_size=8192,
                                    use_unzip=True):

    dest_path = os.path.join(destination_path, dataset_key + ".h5ad.zip")

    with requests.get(dataset_url, stream=True) as r:
        r.raise_for_status()
        if not os.path.isfile(dest_path):
            with open(dest_path, 'wb') as f:
                pbar = tqdm(total=int(r.headers['Content-Length']))
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        pbar.update(len(chunk))

    if use_unzip:
        with zipfile.ZipFile(dest_path, 'r') as zip_ref:
            zip_ref.extractall(destination_path)



from setuptools import setup, find_packages

setup(
    name="pascrd",
    version="0.1.0",
    url='https://github.com/matt-sd-watson/pascrd/',
    project_urls={
        "Issues": "https://github.com/matt-sd-watson/pascrd/issues",
        "Source": "https://github.com/matt-sd-watson/pascrd",
    },
    author="Matthew Watson",
    author_email="mwatson@lunenfeld.ca",
    packages=find_packages(),
    package_dir={"pascrd": "pascrd"},
    package_data={'': ['*.json']},
    include_package_data=True,
    description="",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords=["single-cell download repository"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.9"
    ],
    license="Unlicensed",
    install_requires=["selenium", "requests", "asyncio", "aiohttp"],
    python_requires=">=3.9.0",
)
